import os
import logging

import __main__


GIT_CONTROL = os.system('git rev-parse') == 0

def save_git_sha(job_dir):
    cmd = f"echo $(git rev-parse --short HEAD) > {job_dir}/git_sha.txt"
    os.system(cmd)


def get_pipeline_dir():
    pipeline_dir = os.path.dirname(__main__.__file__)
    if not os.path.exists(pipeline_dir):
        os.makedirs(pipeline_dir, exist_ok=True)
        logging.info(f"Created pipeline directory {pipeline_dir}")

    return pipeline_dir



import os
import logging
import pprint
from uuid import uuid4
import yaml
from pype import utils

class BaseConfig:
    _expected_attributes = {
        "inputs",
        "optional_inputs",
        "params",
        "optional_params",
        "script_path",
        "outputs",
    }

    def __init__(
        self,
        pipeline_dir=None,
        job_id=None,
        inputs=None,
        params=None,
    ):
        """
        Inherit from this config class to create configs for jobs.

        class ExampleConfig(BaseConfig):
            script_path = 'path/to/script/that/configured/by/this/config'
            inputs = {'input_key'}
            params = {'param_key'}
            outputs = {'output_1': 'path/to/where_output_is_saved'}
            optional_params = {'optional_param'}
            optional_inputs = {'optional_input'}

        script pth has to be set so that the config knows where to find the

        To set an optional parameter/input use the default value None.
            outputs has to be a dictionary as it is name of what to save and where to save it
            inputs and params are sets as the user has to specify themselves what there values re
            inputs are always paths to data or outputs from other jobs
            params are the parameters needed to run a job and does not necessarily come from
            outside

        To set an optional parameter/input use the default value None.
        """

        # Verify construcion of child
        assert hasattr(
            self, "script_path"
        ), f"Configuration class for f{type(self)} must include script_path"

        for attr_name in self._expected_attributes:
            if not hasattr(self, attr_name):
                setattr(self, attr_name, set())

        self._verify_attributes()

        # Initialise config
        _verify(inputs, self.inputs, self.optional_inputs)  # pylint: disable=no-member
        _verify(params, self.params, self.optional_params)  # pylint: disable=no-member

        job_id = str(uuid4()) if not job_id else job_id
        if not pipeline_dir:
            pipeline_dir = utils.get_pipeline_dir()
        job_dir = os.path.join(pipeline_dir, job_id)
        output_dir = os.path.join(job_dir, "output")

        outputs = dict()
        for key in self.outputs:  # pylint:disable=no-member
            outputs[key] = os.path.join(
                output_dir, self.outputs[key]  # pylint: disable=no-member
            )

        self.config = {}
        self.config["script_path"] = self.script_path  # pylint: disable=no-member
        self.config["job_id"] = job_id

        self.config["job_dir"] = job_dir

        if inputs:
            self.config["inputs"] = inputs

        if params:
            self.config["params"] = params

        self.config['outputs'] = outputs

        os.makedirs(output_dir, exist_ok=True)
        print(f"Created dirs for {job_id}")

        yaml.dump(self.config, open(os.path.join(job_dir, "config.yaml"), "w"))

        if utils.GIT_CONTROL:
            utils.save_git_sha(job_dir)

    def _verify_attributes(self):
        attrs = set(filter(lambda x: not x.startswith("_"), dir(self)))
        unexpected_attributes = attrs - self._expected_attributes
        if unexpected_attributes:
            raise AttributeError(
                f"Got unexpected attributes {unexpected_attributes}, allowed "
                f"attributes are {self._expected_attributes}"
            )


    def __getitem__(self, key):
        return self.config[key]

    def __str__(self):
        return pprint.pformat(self.__dict__, indent=3)


def _verify(data, expected_keys, optional_keys):
    if not data:
        data = dict()

    missing_keys = expected_keys - data.keys()
    unexpected_keys = data.keys() - expected_keys - optional_keys

    if unexpected_keys:
        raise RuntimeError(f"Got unexpected_keys inputs/params: {unexpected_keys}")
    if missing_keys:
        raise RuntimeError(f"Missing inputs/params: {missing_keys}")



import os

import pytest
import yaml

from pype import BaseConfig


class MockConfig(BaseConfig):
    script_path = "/mock/path"
    params = {"param"}
    optional_params = {"optional_param"}
    inputs = {"input"}
    optional_inputs = {"optional_input"}
    outputs = {"result": "result.ext"}


params = {"param": 1, "optional_param": 2}
inputs = {"input": 1, "optional_input": 2}


def test_mockconfig_can_instantiate(tmpdir):
    MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)


def test_mockconfig_can_has_attributes(tmpdir):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)

    for key in ["inputs", "params", "outputs", "job_id"]:
        assert key in config.config


def test_mockconfig_can_instantiate_twice(tmpdir):
    MockConfig(job_id="foo", pipeline_dir=tmpdir, params=params, inputs=inputs)
    config = MockConfig(job_id="bar", pipeline_dir=tmpdir, params=params, inputs=inputs)

    assert config["outputs"]["result"].endswith("bar/output/result.ext")


def test_mock_config_creates_the_right_paths(tmpdir):
    config = MockConfig(job_id="bla", pipeline_dir=tmpdir, params=params, inputs=inputs)

    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "output"))
    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "git_sha.txt"))
    assert os.path.exists(os.path.join(tmpdir, config["job_id"], "config.yaml"))


def test_cant_instantiate_when_missing_input(tmpdir):
    with pytest.raises(RuntimeError):
        MockConfig(params=params, pipeline_dir=tmpdir, inputs=dict())


def test_cant_instantiate_when_missing_param(tmpdir):
    with pytest.raises(RuntimeError):
        MockConfig(params=dict(), inputs=inputs, pipeline_dir=tmpdir)


def test_cant_instantiate_with_unexpected_input(tmpdir):
    with pytest.raises(RuntimeError):
        inputs.update({"unexpected_input": "2"})
        MockConfig(params=params, inputs=inputs, pipeline_dir=tmpdir)


def test_cant_instantiate_with_unexpected_param(tmpdir):
    with pytest.raises(RuntimeError):
        params.update({"unexpected_param": "2"})
        MockConfig(params=params, inputs=inputs, pipeline_dir=tmpdir)


def test_cant_create_class_with_wrong_attributes(tmpdir):
    class WrongConfig(BaseConfig):
        script_path = "script/path"
        wrong_name_for_inputs = {"hello"}

    with pytest.raises(AttributeError):
        WrongConfig(pipeline_dir=tmpdir)


# def mock_job(config):
#     f = open(config["output_paths"]["result"], "w")
#     f.write(config["params"]["string"])
#
#
# def test_integration(tmpdir):
#     params = {"string": "Hello World!"}
#     config = MockConfig(pipeline_dir=tmpdir, job_id="bla", params=params)
#     config.save()
#
#     loaded_config = yaml.load(
#         open(os.path.join(config.output_dir, "config.yaml")), Loader=yaml.FullLoader
#     )
#
#     mock_job(loaded_config)
#
# def test_script_path_in_dict(tmpdir):
#     params = {"string": "Hello World!"}
#     config = MockConfig(pipeline_dir=tmpdir, job_id="bla", params=params)
#     assert "script_path" in config.__dict__
#
#
# def test_config_throws_error_without_script_path(tmpdir):
#     class MockConfigWithoutScriptPath(BaseConfig):
#         pass
#
#     with pytest.raises(Exception):
#         _ = MockConfigWithoutScriptPath(tmpdir)
#
# class MockConfigDefault(BaseConfig):
#     default_params = {"default_param": 1}
#     default_inputs = {"default_input": 2}
#     outputs = {"result": "result"}
#     script_path = "/bla/bla"
#
#
# def test_instantiate_default_config_without_values(tmpdir):
#     mockconfig = MockConfigDefault(pipeline_dir=tmpdir, job_id="bla", )
#
#     assert "default_param" in mockconfig.params
#     assert mockconfig.params["default_param"] == 1
#     assert "default_input" in mockconfig.inputs
#     assert mockconfig.inputs["default_input"] == 2
#
#
# def test_instantiate_default_config_with_values(tmpdir):
#     params = {"default_param": 3}
#     inputs = {"default_input": 4}
#     mockconfig = MockConfigDefault(pipeline_dir=tmpdir, job_id="bla", params=params, inputs=inputs)
#
#     assert "default_param" in mockconfig.params
#     assert mockconfig.params["default_param"] == 3
#     assert "default_input" in mockconfig.inputs
#     assert mockconfig.inputs["default_input"] == 4



from setuptools import setup, find_packages

setup(
    author="Mathias Schreiner",
    author_email="matschreiner@gmail.com",
    url="https://gitlab.com/matschreiner/pype",
    download_url="https://gitlab.com/matschreiner/pype/-/archive/v1.0.0/pype-v1.0.0.tar.gz",
    name="pype-ms",
    version="1.0.0",
    packages=find_packages(),
    install_requires=[
        "pyyaml==5.4",
        "pytest",
        "Click",
    ],
    entry_points={
        "console_scripts": [
            "pype = pype.cli:cli",
        ],
    },
)



import datetime
import importlib
import logging
import os
import shutil
import socket
import sys
import traceback

import click
import yaml

from pype import utils


logging.basicConfig(level=logging.INFO)


class Status:
    def __init__(self, dir_):
        self.dir = dir_
        self.status = "status"
        for status in ["Done", "Failed", "Running"]:
            if os.path.exists(os.path.join(dir_, status)):
                self.status = status

        self.status_path = os.path.join(dir_, self.status)

        with open(self.status_path, "w") as f:
            f.write("Host: " + socket.gethostname() + "\n")

    def done(self):
        self._set_status("Done")

    def running(self):
        self._set_status("Running")

    def failed(self):
        self._set_status("Failed")

    def _set_status(self, status):
        timestamp = datetime.datetime.now().strftime("%y/%m/%d-%H:%M:%S")

        with open(self.status_path, "a") as status_file:
            status_file.write(status + ": " + timestamp + "\n")

        status_path = os.path.join(self.dir, status)
        os.rename(self.status_path, status_path)

        self.status = status
        self.status_path = status_path


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--rerun",
    default=False,
    is_flag=True,
    help="Delete everything in the output dir of the job and rerun the job with the specified "
    "config.yaml",
)
@click.option(
    "--allow-uncommitted",
    "-au",
    default=False,
    is_flag=True,
    help="Allow to run job even if there are uncommitted changes in the directory",
)
@click.option(
    "--log",
    "-l",
    default=False,
    is_flag=True,
    help="will log stdout and stderr to a file log.txt instead of stdout, useful when using "
    "screen for example",
)
@click.argument("config_file")
def run(config_file, rerun, allow_uncommitted, log):
    config = yaml.load(open(config_file, "r"), Loader=yaml.FullLoader)
    job_dir = config["job_dir"]

    if rerun:
        _remove_all_but_config(job_dir)

    status = Status(job_dir)

    if status.status == "Running" or status.status == "Done":
        if permission_to_continue(f"Job is {status.status}."):
            print("job aborted")
            sys.exit()

    if utils.GIT_CONTROL:
        if not allow_uncommitted >= _uncomitted():
            if permission_to_continue("You have uncomitted changes."):
                status.failed()
                logging.info("Job aborted.")
                sys.exit()

        utils.save_git_sha(job_dir)

    status.running()

    if log:
        sys.stdout = open(os.path.join(job_dir, "logs.txt"), "w")

    try:
        module = _import_module(config["script_path"])
        if not hasattr(module, "main"):
            raise RuntimeError(f"{config['script_path']} has no main function.")

        module.main(config)
        status.done()

    except Exception:  # pylint: disable=broad-except
        status.failed()
        print(traceback.format_exc())


def permission_to_continue(msg):
    return input(msg + "Type 'y' or 'yes' to continue anyways\n").lower() not in [
        "y",
        "yes",
    ]


def _remove_all_but_config(job_dir):
    for f in os.listdir(job_dir):
        if f == "config.yaml":
            continue

        full_path = os.path.join(job_dir, f)

        if os.path.isdir(full_path):
            shutil.rmtree(full_path)
        else:
            os.remove(full_path)


def _import_module(path):
    spec = importlib.util.spec_from_file_location("module.name", path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def _uncomitted():
    cmd = r"git status | grep -q '\smodified:\s'"
    code = os.system(cmd)
    return code == 0
